import re
import time
from math import ceil

from pkg_resources import resource_string

from .exceptions import RateLimitError


class NoRateLimitBucket:
    def take(self):
        return

    def update(self, values):
        return


class ResourceRateLimitBucket:
    def __init__(self, bucket_id):
        self.bucket_id = bucket_id
        self.remaining = None
        self.reset = None

    def take(self):
        if (
            self.remaining is not None
            and self.remaining <= 0
            and time.time() < self.reset
        ):
            raise RateLimitError(self.reset)
        self.remaining -= 1

    def update(self, values):
        self.remaining = int(values["X-RateLimit-Remaining"])
        self.reset = int(values["X-RateLimit-Reset"])


class GlobalRateLimitBucket:
    def __init__(self):
        self.is_ratelimited = False
        self.reset = None

    def take(self):
        if self.is_ratelimited and time.time() < self.reset:
            raise RateLimitError(self.reset, is_global=True)

    def update(self, values):
        self.is_ratelimited = (
            values.get("X-RateLimit-Global", "false").lower() == "true"
        )
        if self.is_ratelimited:
            retry_after = ceil(int(values.get("Retry-After", 0)) / 1000)
            self.reset = time.time() + retry_after


class RateLimiter:
    no_ratelimit_bucket = NoRateLimitBucket()

    def __init__(self):
        # bucket id to bucket mapping
        self.buckets = {}
        # resource to bucket mapping
        self.resource_buckets = {}
        self.global_bucket = GlobalRateLimitBucket()

    def on_request(self, method, path):
        self.global_bucket.take()
        self.get_bucket(method, path).take()

    def on_response(self, method, path, headers, status_code):
        bucket = None
        if headers.get("X-RateLimit-Global"):
            bucket = self.global_bucket
        else:
            bucket_id = headers.get("X-RateLimit-Bucket")
            bucket = self.get_bucket(method, path, bucket_id)
        bucket.update(headers)

        if status_code == 429:
            raise RateLimitError(bucket.reset, is_global=bucket is self.global_bucket)

    def get_bucket(self, method, path, bucket_id=None):
        resource = get_resource(path)
        key = (method, resource)
        try:
            bucket = self.resource_buckets[key]
            if bucket_id is None or bucket_id == bucket.bucket_id:
                return bucket
        except KeyError:
            pass

        if not bucket_id:
            return self.no_ratelimit_bucket

        try:
            bucket = self.buckets[bucket_id]
        except KeyError:
            bucket = self.buckets[bucket_id] = ResourceRateLimitBucket(bucket_id)
        self.resource_buckets[key] = bucket
        return bucket


def extract_patterns(mappings):
    resources_patterns = []

    for mapping in mappings:
        mapping = mapping.strip()
        if not mapping or mapping.startswith("#"):
            continue

        pattern, resource = mapping.split("=")
        pattern = re.compile(pattern)
        resources_patterns.append((pattern, resource))

    return resources_patterns


mappings = resource_string("smalld.resources", "ratelimit_buckets").decode("utf-8")
mappings = extract_patterns(mappings.split("\n"))


def get_resource(path):
    path = path.strip().strip("/")
    for (pattern, template) in mappings:
        match = pattern.fullmatch(path)
        if not match:
            continue
        return match.expand(template)
    return path



import logging
import sys
import time
from threading import Event, Thread

from .exceptions import NetworkError

logger = logging.getLogger("smalld")


def add_standard_listeners(smalld):
    sequence = SequenceNumber(smalld)
    Identify(smalld, sequence)
    Heartbeat(smalld, sequence)


OP_HEARTBEAT = 1
OP_IDENTIFY = 2
OP_INVALID_SESSION = 9
OP_RESUME = 6
OP_RECONNECT = 7
OP_HELLO = 10
OP_HEARTBEAT_ACK = 11


class SequenceNumber:
    def __init__(self, smalld):
        self.smalld = smalld
        self.number = None

        smalld.on_gateway_payload()(self.on_payload)

    def on_payload(self, data):
        if data.s:
            self.number = data.s


class Identify:
    def __init__(self, smalld, sequence):
        self.smalld = smalld
        self.sequence = sequence
        self.session_id = None

        smalld.on_ready()(self.on_ready)
        smalld.on_resumed()(self.on_resumed)
        smalld.on_gateway_payload(op=OP_HELLO)(self.on_hello)
        smalld.on_gateway_payload(op=OP_INVALID_SESSION)(self.on_invalid_session)
        smalld.on_gateway_payload(op=OP_RECONNECT)(self.on_reconnect)

    def on_ready(self, data):
        logger.info("Ready.")
        self.session_id = data.session_id

    def on_resumed(self, data):
        logger.info("Resumed.")

    def on_hello(self, data):
        if self.session_id and self.sequence.number:
            self.resume()
        else:
            self.identify()

    def on_invalid_session(self, data):
        self.session_id = None
        time.sleep(2)
        self.identify()

    def on_reconnect(self, data):
        self.smalld.reconnect()

    def identify(self):
        logger.info("Identifying...")
        self.smalld.send_gateway_payload(
            {
                "op": OP_IDENTIFY,
                "d": {
                    "token": self.smalld.token,
                    "properties": {
                        "$os": sys.platform,
                        "$browser": "smalld.py",
                        "$device": "smalld.py",
                    },
                    "compress": False,
                    "intents": self.smalld.intents.value,
                },
            }
        )

    def resume(self):
        logger.info("Resuming...")
        self.smalld.send_gateway_payload(
            {
                "op": OP_RESUME,
                "d": {
                    "token": self.smalld.token,
                    "session_id": self.session_id,
                    "seq": self.sequence.number,
                },
            }
        )


class Heartbeat:
    def __init__(self, smalld, sequence):
        self.smalld = smalld
        self.sequence = sequence
        self.thread = None
        self.heartbeat_interval = None
        self.received_ack = Event()

        smalld.on_gateway_payload(op=OP_HELLO)(self.on_hello)
        smalld.on_gateway_payload(op=OP_HEARTBEAT)(self.on_heartbeat)
        smalld.on_gateway_payload(op=OP_HEARTBEAT_ACK)(self.on_heartbeat_ack)

    def on_hello(self, data):
        self.heartbeat_interval = data.d.heartbeat_interval

        if not self.thread or not self.thread.is_alive():
            self.thread = Thread(target=self.run_heartbeat_loop)
            self.thread.start()

    def on_heartbeat(self, data):
        self.send_heartbeat()

    def on_heartbeat_ack(self, data):
        self.received_ack.set()

    def run_heartbeat_loop(self):
        interval = self.heartbeat_interval / 1000
        time.sleep(interval)
        while not self.smalld.closed:
            try:
                self.send_heartbeat()
            except NetworkError:
                continue
            finally:
                time.sleep(interval)

            if self.received_ack.is_set():
                self.received_ack.clear()
            else:
                self.smalld.reconnect()
                break

    def send_heartbeat(self):
        self.smalld.send_gateway_payload(
            {"op": OP_HEARTBEAT, "d": self.sequence.number}
        )



from unittest.mock import patch

import pytest
import responses
from smalld import HttpError, NetworkError
from smalld.smalld import HttpClient


@pytest.fixture(autouse=True)
def limiter():
    with patch("smalld.smalld.RateLimiter") as ratelimiter_cls:
        yield ratelimiter_cls.return_value


@responses.activate
@pytest.mark.parametrize(
    "method, path, client_method",
    [
        ("GET", "/get", HttpClient.get),
        ("POST", "/post", HttpClient.post),
        ("PUT", "/put", HttpClient.put),
        ("PATCH", "/patch", HttpClient.patch),
        ("DELETE", "/delete", HttpClient.delete),
    ],
)
def test_httpclient_returns_successful_responses(method, path, client_method):
    url = f"https://domain.com/{path}"
    responses.add(method, url, json={"data": "some data"}, status=200)
    client = HttpClient("token", "https://domain.com")

    res = client_method(client, path)

    assert hasattr(res, "data") and res.data == "some data"
    assert len(responses.calls) == 1
    assert responses.calls[0].request.url == url
    assert responses.calls[0].request.method == method


@responses.activate
def test_httpclient_raises_for_non_2xx_status():
    statuses = [400, 403, 404, 500, 501, 502]  # sample non 2xx statuses
    client = HttpClient("token", "https://domain.com")

    for status in statuses:
        responses.add(responses.GET, "https://domain.com/get", status=status)
        with pytest.raises(HttpError):
            res = client.send_request("GET", "get")
        responses.remove(responses.GET, "https://domain.com/get")


@responses.activate
def test_httpclient_raises_for_connection_errors():
    client = HttpClient("token", "https://domain.com")

    with pytest.raises(NetworkError):
        client.get("path")


@responses.activate
def test_httpclient_calls_limiter_before_and_after_request(limiter):
    def assert_called_before_request(*args, **kwargs):
        assert len(responses.calls) == 0

    limiter.on_request.side_effect = assert_called_before_request

    headers = {"X-Header": "value"}
    responses.add(
        responses.GET, "https://domain.com/get", json={"data": "value"}, headers=headers
    )
    extra_headers = {"Content-Type": "application/json"}  # added by responses

    client = HttpClient("token", "https://domain.com")
    client.get("get")

    limiter.on_request.assert_called_once_with("GET", "get")
    limiter.on_response.assert_called_once_with(
        "GET", "get", dict(**headers, **extra_headers), 200
    )
    assert len(responses.calls) == 1


@responses.activate
def test_httpclient_handles_no_content():
    responses.add(responses.GET, "https://domain.com/get", status=204)
    client = HttpClient("token", "https://domain.com")
    res = client.get("get")
    assert res == {}


@responses.activate
@pytest.mark.parametrize("invalid_json", ["", "invalid json", '{"key": value}'])
def test_httpclient_raises_for_response_decoding_errors(invalid_json):
    responses.add(responses.GET, "https://domain.com/get", body=invalid_json)

    client = HttpClient("token", "https://domain.com")
    with pytest.raises(HttpError):
        client.get("get")



class SmallDException(Exception):
    pass


class HttpError(SmallDException):
    def __init__(self, *args, response=None, **kwargs):
        self.response = response
        super().__init__(*args, **kwargs)


class NetworkError(SmallDException):
    pass


class RateLimitError(SmallDException):
    def __init__(self, reset, *, is_global=False):
        super().__init__(f"rate limited until {reset}")
        self.reset = reset
        self.is_global = is_global



from unittest import mock

import pytest
from smalld.gateway import Gateway
from websocket import ABNF, WebSocketException

CONNECTION_URL = "ws://example.url/"


@pytest.fixture()
def ws_mock():
    with mock.patch("smalld.gateway.WebSocket", autospec=True) as ws_class_mock:
        instance = ws_class_mock.return_value
        instance.readlock = mock.MagicMock()
        instance.connected = True

        yield instance


def test_gateway_connects_on_iteration(ws_mock):
    ws_mock.recv_data.return_value = ABNF.OPCODE_TEXT, b"{}"
    gateway = Gateway(CONNECTION_URL)

    it = iter(gateway)
    next(it)
    it.close()

    ws_mock.connect.assert_called_once_with(CONNECTION_URL)


def test_gateway_yields_decoded_attrdict(ws_mock):
    test_inputs = [(ABNF.OPCODE_TEXT, b"{}"), (ABNF.OPCODE_TEXT, b'{"key": "value"}')]
    expected_results = [{}, {"key": "value"}]
    ws_mock.recv_data.side_effect = test_inputs
    gateway = Gateway(CONNECTION_URL)

    it = iter(gateway)
    results = [next(it), next(it)]
    it.close()

    assert len(ws_mock.recv_data.mock_calls) == len(test_inputs)
    assert results == expected_results
    assert hasattr(results[1], "key")


def test_gateway_skips_empty_data(ws_mock):
    test_input, expected = (ABNF.OPCODE_TEXT, b"{}"), {}
    ws_mock.recv_data.side_effect = [(ABNF.OPCODE_TEXT, b""), test_input]
    gateway = Gateway(CONNECTION_URL)

    it = iter(gateway)
    result = next(it)
    it.close()

    assert len(ws_mock.recv_data.mock_calls) == 2
    assert result == expected


def test_gateway_closes_on_close_event(ws_mock):
    expected_code, expected_reason = 16, "Reason"
    ws_mock.recv_data.return_value = ABNF.OPCODE_CLOSE, b"\x00\x10Reason"
    gateway = Gateway(CONNECTION_URL)

    for data in gateway:
        pytest.fail("No data should be emitted")

    assert gateway.close_reason
    assert gateway.close_reason.code == expected_code
    assert gateway.close_reason.reason == expected_reason


def test_gateway_end_when_websocket_exception(ws_mock):
    ws_mock.recv_data.side_effect = [WebSocketException()]

    gateway = Gateway(CONNECTION_URL)

    for data in gateway:
        pytest.fail("Should receive no data")



from smalld import SmallD

smalld = SmallD()


@smalld.on_message_create()
def on_message(msg):
    if msg.content == "++ping":
        smalld.post(f"/channels/{msg.channel_id}/messages", {"content": "pong"})


smalld.run()



from unittest.mock import patch

import pytest
from smalld.ratelimit import *


class ControllableTime:
    def __init__(self):
        self.time = 0

    def __call__(self):
        return self.time

    def set_to(self, time):
        self.time = time


def make_ratelimit_headers(
    bucket="default", limit=10, remaining=1, reset=1, reset_after=1
):
    return {
        "X-RateLimit-Limit": str(limit),
        "X-RateLimit-Remaining": str(remaining),
        "X-RateLimit-Reset": str(reset),
        "X-RateLimit-Reset-After": str(reset_after),
        "X-RateLimit-Bucket": bucket,
    }


def make_global_ratelimit_headers(retry_after):
    return {"Retry-After": str(retry_after), "X-RateLimit-Global": "true"}


@pytest.fixture(autouse=True)
def time():
    with patch("time.time") as time_mock:
        time = ControllableTime()
        time_mock.side_effect = time
        yield time


def test_resource_ratelimit_bucket(time):
    time.set_to(1000)
    bucket = ResourceRateLimitBucket("abc123")
    bucket.update(make_ratelimit_headers(reset=1001, reset_after=1))

    assert bucket.bucket_id == "abc123"
    assert bucket.reset == 1001
    assert bucket.remaining == 1

    bucket.take()  # doesn't raise
    assert bucket.remaining == 0

    with pytest.raises(RateLimitError) as exc_info:
        bucket.take()

    e = exc_info.value
    assert e.reset == 1001 and e.is_global == False


def test_global_ratelimit_bucket(time):
    time.set_to(100)
    limit = GlobalRateLimitBucket()

    assert not limit.is_ratelimited
    limit.take()  # doesn't raise

    limit.update(make_global_ratelimit_headers(100))

    assert limit.is_ratelimited

    with pytest.raises(RateLimitError) as exc_info:
        limit.take()

    e = exc_info.value
    assert e.reset == 101 and e.is_global == True


def test_ratelimit_passes_first_request():
    limiter = RateLimiter()
    limiter.on_request("GET", "/path/to/resource")  # doesn't raise


def test_ratelimit_passes_good_response(time):
    limiter = RateLimiter()
    limiter.on_response(
        "GET", "/path/to/resource", make_ratelimit_headers(), 200
    )  # doesn't raise


@pytest.mark.parametrize(
    "start, reset, is_global, response",
    [
        (
            0,
            1,
            True,
            ("GET", "/path/to/resource", make_global_ratelimit_headers(100), 429),
        ),
        (
            1000,
            1001,
            False,
            (
                "GET",
                "/path/to/resource",
                make_ratelimit_headers("abc123", 10, 0, 1001, 1),
                429,
            ),
        ),
    ],
)
def test_ratelimit_raises_on_limit_exhausted_response(
    time, start, reset, is_global, response
):
    time.set_to(start)
    limiter = RateLimiter()
    with pytest.raises(RateLimitError) as exc_info:
        limiter.on_response(*response)

    e = exc_info.value
    assert e.reset == reset and e.is_global == is_global


def test_ratelimit_raises_on_request_exhausted_resource(time):
    time.set_to(1000)
    limiter = RateLimiter()
    bucket = limiter.resource_buckets[
        ("GET", "path/to/resource")
    ] = ResourceRateLimitBucket("abc123")
    bucket.update(make_ratelimit_headers("abc123", 10, 0, 1002, 2))

    with pytest.raises(RateLimitError) as exc_info:
        limiter.on_request("GET", "path/to/resource")

    assert exc_info.value.reset == 1002


@pytest.mark.parametrize(
    "path, resource",
    [
        ("channels/2909267986263572999", "channels/2909267986263572999"),
        ("guilds/197038439483310086", "guilds/197038439483310086"),
        ("webhooks/223704706495545344", "webhooks/223704706495545344"),
        ("/channels/2909267986263572999/", "channels/2909267986263572999"),
        (
            "/guilds/197038439483310086/members/63269852323648",
            "guilds/197038439483310086/members/{user.id}",
        ),
        ("/users/9864325349523", "users/{user.id}"),
        ("/users/@me/guilds", "users/@me/guilds"),
        ("/invites/0vCdhLbwjZZTWZLD", "invites/{invite.code}"),
        ("/unknown/path", "unknown/path"),
    ],
)
def test_get_resource(path, resource):
    assert get_resource(path) == resource



import json

from attrdict import AttrDict
from websocket import (
    ABNF,
    WebSocket,
    WebSocketConnectionClosedException,
    WebSocketException,
)

from .exceptions import NetworkError


class CloseReason:
    def __init__(self, code, reason):
        self.code = code
        self.reason = reason

    def __str__(self):
        return f"{code}: {reason}"

    @staticmethod
    def parse(data):
        code = int.from_bytes(data[:2], "big")
        reason = data[2:].decode("utf-8")
        return CloseReason(code, reason)


class Gateway:
    def __init__(self, url):
        self.url = url
        self.ws = WebSocket()
        self.close_reason = None

    def __iter__(self):
        try:
            self.ws.connect(self.url)
        except WebSocketException:
            raise ConnectionError

        while self.ws.connected:
            try:
                with self.ws.readlock:
                    opcode, data = self.ws.recv_data()
            except WebSocketException:
                return

            if data and opcode == ABNF.OPCODE_CLOSE:
                self.close_reason = CloseReason.parse(data)
                return

            if data and opcode == ABNF.OPCODE_TEXT:
                decoded_data = data.decode("utf-8")
                yield AttrDict(json.loads(decoded_data))

    def send(self, data):
        try:
            self.ws.send(data)
        except WebSocketConnectionClosedException:
            raise NetworkError

    def close(self):
        self.ws.close()



import json
import logging
import os
import time
from enum import Flag
from threading import Event

from pkg_resources import get_distribution

import requests
from attrdict import AttrDict

from .exceptions import HttpError, NetworkError
from .gateway import Gateway
from .ratelimit import RateLimiter
from .standard_listeners import add_standard_listeners

logger = logging.getLogger("smalld")


__version__ = get_distribution("smalld").version


class Intent(Flag):
    GUILDS = 1 << 0
    GUILD_MEMBERS = 1 << 1
    GUILD_BANS = 1 << 2
    GUILD_EMOJIS = 1 << 3
    GUILD_INTEGRATIONS = 1 << 4
    GUILD_WEBHOOKS = 1 << 5
    GUILD_INVITES = 1 << 6
    GUILD_VOICE_STATES = 1 << 7
    GUILD_PRESENCES = 1 << 8
    GUILD_MESSAGES = 1 << 9
    GUILD_MESSAGE_REACTIONS = 1 << 10
    GUILD_MESSAGE_TYPING = 1 << 11
    DIRECT_MESSAGES = 1 << 12
    DIRECT_MESSAGE_REACTIONS = 1 << 13
    DIRECT_MESSAGE_TYPING = 1 << 14

    @staticmethod
    def all():
        return ~Intent(0)

    @staticmethod
    def unprivileged():
        return ~(Intent.GUILD_PRESENCES | Intent.GUILD_MEMBERS)


recoverable_error_codes = {
    *range(1000, 1016),  # standard protocol error codes
    4000,  # unknown error
    4001,  # unknown opcode
    4002,  # decode error
    4005,  # already authenticated
    4007,  # invalid sequence
    4009,  # session timed out
}


class SmallD:
    def __init__(
        self,
        token=os.environ.get("SMALLD_TOKEN"),
        base_url="https://discord.com/api/v6",
        intents=Intent.unprivileged(),
    ):
        if not token:
            raise ValueError("No bot token provided")

        self.token = token
        self.base_url = base_url
        self.intents = intents
        self.listeners = []
        self.closed_event = Event()

        self.http = HttpClient(token, base_url)

        self.get = self.http.get
        self.post = self.http.post
        self.put = self.http.put
        self.patch = self.http.patch
        self.delete = self.http.delete

        add_standard_listeners(self)

    def __getattr__(self, name):
        if name.startswith("on_"):
            return lambda: self.on_dispatch(t=name.strip("on_").upper())

        super().__getattr__(name)

    def on_dispatch(self, t):
        def decorator(f):
            self.on_gateway_payload(op=0, t=t)(lambda payload: f(payload.d))

        return decorator

    def on_gateway_payload(self, op=None, t=None):
        def decorator(f):
            def filtered_payload_listener(data):
                if op and data.op != op:
                    return

                if t and data.t != t:
                    return

                f(data)

            self.listeners.append(filtered_payload_listener)

        return decorator

    def send_gateway_payload(self, data):
        payload = json.dumps(data)
        logger.debug("gateway payload sent: %s", payload)
        self.gateway.send(payload)

    @property
    def closed(self):
        return self.closed_event.is_set()

    def reconnect(self):
        self.gateway.close()

    def close(self):
        self.closed_event.set()
        self.http.close()
        self.reconnect()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if not self.closed:
            self.close()

    def run(self):
        logger.info("Running (SmallD v%s)...", __version__)

        self.closed_event.clear()

        while not self.closed:
            logger.info("Gateway connecting...")
            gateway_url = self.get("/gateway/bot").url

            self.gateway = Gateway(gateway_url)

            for data in self.gateway:
                logger.debug("gateway payload received: %s", data)
                for listener in self.listeners:
                    listener(data)

            reason = self.gateway.close_reason
            if reason and not reason.code in recoverable_error_codes:
                logger.fatal(
                    "Unrecoverable gateway closure: %s", self.gateway.close_reason
                )
                self.close()

            if not self.closed:
                logger.info("Pausing before reconnect...")
                time.sleep(5)


class HttpClient:
    def __init__(self, token, base_url):
        self.token = token
        self.base_url = base_url
        self.session = requests.Session()
        self.session.headers.update(self.headers())
        self.limiter = RateLimiter()

    def headers(self):
        return {
            "Authorization": f"Bot {self.token}",
            "User-Agent": f"DiscordBot (https://github.com/princesslana/smalld.py, {__version__})",
        }

    def get(self, *args, **kwargs):
        return self.send_request("GET", *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.send_request("POST", *args, **kwargs)

    def put(self, *args, **kwargs):
        return self.send_request("PUT", *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self.send_request("PATCH", *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self.send_request("DELETE", *args, **kwargs)

    def send_request(self, method, path, payload="", attachments=None, params=None):
        if attachments:
            files = [(f"file{idx}", a) for idx, a in enumerate(attachments)]
            args = {"data": {"payload_json": json.dumps(payload)}, "files": files}
        elif payload:
            args = {"json": payload}
        else:
            args = {}

        if params:
            args["params"] = params

        self.limiter.on_request(method, path)

        try:
            res = self.session.request(method, f"{self.base_url}/{path}", **args)
        except (requests.ConnectionError, requests.Timeout):
            raise NetworkError
        except requests.RequestException:
            raise HttpError

        self.limiter.on_response(method, path, res.headers, res.status_code)

        if not res.ok:
            raise HttpError(response=res)

        try:
            content = res.json() if res.status_code != 204 else {}
        except json.JSONDecodeError:
            raise HttpError(response=res)

        return AttrDict(content)

    def close(self):
        self.session.close()



from unittest.mock import Mock, patch

import pytest
from attrdict import AttrDict
from smalld.gateway import CloseReason
from smalld.smalld import SmallD, recoverable_error_codes


def prepare_gateway_mock(
    gateway_mock, smalld, side_effects=[()], close_reason=None, auto_close=True
):
    """Configures __iter__ for gateway.

    Each call to __iter__ is represented by an iterable in the side_effects list.
    If the iterable yields an exception, the exception is raised and iteration halts.

    By default this configuration:
      - closes gracefully
      - iterates once
    """

    side_effects = list(side_effects)
    finished = False

    def iter_gateway():
        nonlocal finished
        assert not finished, "infinite loop prevention"

        if not side_effects:
            finished = True
            if auto_close:
                smalld.close()
            return

        values = side_effects.pop(0)
        for value in values:
            if (
                isinstance(value, BaseException)
                or isinstance(value, type)
                and issubclass(value, BaseException)
            ):
                raise value
            yield AttrDict(value)

        if close_reason:
            gateway_mock.close_reason.return_value = CloseReason(*close_reason)

    gateway_mock.__iter__.side_effect = iter_gateway


@pytest.fixture(autouse=True)
def gateway_mock():
    with patch("smalld.smalld.Gateway") as gateway_cls:
        yield gateway_cls.return_value


@pytest.fixture(autouse=True)
def client_mock():
    with patch("smalld.smalld.HttpClient", autospec=True) as client_cls:
        instance = client_cls.return_value
        # for get gateway request
        instance.get.return_value = AttrDict({"url": "url/to/gateway"})
        yield instance


@pytest.fixture(autouse=True)
def sleep_mock():
    with patch("time.sleep") as sleep_mock:
        yield sleep_mock


test_data = [
    {"op": 1, "t": None, "d": {"key1": "value1"}, "s": 0},
    {"op": 0, "t": "EVENT", "d": {"key1", "value1"}, "s": 0},
]


@pytest.mark.parametrize(
    "payload_type, expected_data",
    [({"op": 1}, test_data[0]), ({"t": "EVENT"}, test_data[1])],
)
def test_smalld_calls_listener_on_payload(payload_type, expected_data, gateway_mock):
    callback = Mock()
    smalld = SmallD("token")
    prepare_gateway_mock(gateway_mock, smalld, [test_data])

    smalld.on_gateway_payload(**payload_type)(callback)
    smalld.run()

    callback.assert_called_once_with(expected_data)


def test_smalld_calls_event_listener_on_payload(gateway_mock):
    data = {"key", "value"}
    payload = {"op": 0, "t": "CREATE_MESSAGE", "d": data, "s": 0}
    callback = Mock()
    smalld = SmallD("token")
    prepare_gateway_mock(gateway_mock, smalld, [[payload]])

    smalld.on_create_message()(callback)
    smalld.run()

    callback.assert_called_once_with(data)


def test_smalld_ends_for_non_recoverable_gateway_errors(gateway_mock):
    smalld = SmallD("token")

    assert -1 not in recoverable_error_codes  # sanity check
    prepare_gateway_mock(gateway_mock, smalld, close_reason=(-1, "reason"))

    smalld.run()


@pytest.mark.parametrize("code", list(recoverable_error_codes))
def test_smalld_handles_recoverable_error(code, gateway_mock):
    smalld = SmallD("token")
    prepare_gateway_mock(gateway_mock, smalld, close_reason=(code, "reason"))

    smalld.run()  # doesn't raise


def test_smalld_closes_properly(gateway_mock, client_mock):
    smalld = SmallD("token")
    prepare_gateway_mock(gateway_mock, smalld)

    with smalld:
        smalld.run()

    assert smalld.closed is True
    gateway_mock.close.assert_called_once()
    client_mock.close.assert_called_once()


def test_smalld_closes_properly_with_exception(gateway_mock, client_mock):
    smalld = SmallD("token")
    exception = Exception()
    prepare_gateway_mock(gateway_mock, smalld, [[exception]], auto_close=False)

    with pytest.raises(Exception) as exc_info:
        with smalld:
            smalld.run()

    assert exc_info.value is exception
    assert smalld.closed is True
    gateway_mock.close.assert_called_once()
    client_mock.close.assert_called_once()



from io import BytesIO

import requests
from smalld import SmallD

smalld = SmallD()

CAT_API = "http://aws.random.cat/meow"


@smalld.on_message_create()
def on_message(msg):
    if msg.content == "++cat":
        cat_url = requests.get(CAT_API).json()["file"]
        cat_pic = requests.get(cat_url).content
        smalld.post(
            f"/channels/{msg.channel_id}/messages",
            payload={"content": "Enjoy your cat pic!"},
            attachments=[("cat.jpg", BytesIO(cat_pic), "image/jpeg")],
        )


smalld.run()



import os

from setuptools import setup


def readme():
    readme_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md")

    with open(readme_path, encoding="utf-8") as f:
        return f.read()


setup(
    name="smalld",
    description="A minimalist python client for the Discord API",
    long_description=readme(),
    long_description_content_type="text/markdown",
    packages=["smalld", "smalld.resources"],
    package_data={"smalld.resources": ["*"]},
    use_scm_version=True,
    install_requires=[
        "attrdict>=2.0.1",
        "requests>=2.23.0",
        "websocket_client>=0.57.0",
    ],
    setup_requires=["setuptools-scm==3.3.3"],
)
